#include <string.h>
#include "ets_sys.h"
#include "osapi.h"
#include "gpio.h"
#include "os_type.h"
#include "user_config.h"
#include "driver/uart.h"
#include "user_interface.h"
#include "espconn.h"
#include "user_tcp.h"
#include <mem.h>

#define NETWORK_UPDATE_RATE 5

#define user_procTaskQueueLen    10

#if QUAD_NUM == 0
uint8_t mac[6] = {0xb8, 0xd7, 0x63, 0x00, 0xfe, 0xef};
#elif QUAD_NUM == 1
uint8_t mac[6] = {0xb8, 0xd7, 0x63, 0x00, 0xfe, 0xf0};
#else
uint8_t mac[6] = {};
#endif

static uint8 *_rxBuffer;
static uint8 *_txBuffer;

os_event_t    user_procTaskQueue[user_procTaskQueueLen];
static void uart_task(os_event_t *events);
static void tcp_recvHandler(uint16 len);

static volatile os_timer_t networkTimer;
static volatile uint8 _ledSet;

static void wifi_handler(System_Event_t *event);

static volatile int uartCount = 0;
static volatile uint8 _uartTxFlag;

void network_task(void *arg) {
    static uint8_t ledState = 0;
    
    if(_ledSet) {
        ledState = 1;
        _ledSet = 0;

        //Turn on activity LED
        gpio_output_set(BIT0, 0, BIT0, 0);
    }

    if(ledState) {
        if(ledState++ > 10) {
            ledState = 0;
            gpio_output_set(0, BIT0, BIT0, 0);
        }
    }

#if DEBUG_NETWORK_TASK == 1
    static int i = 0;
    if(i == 0) {
        i = 100;

        char msg[128];
        os_sprintf(msg, "[Heap] %d\r\n", 
            (int)system_get_free_heap_size());
        uart_debugSend(msg);
    }
    i--;
#endif
}

//Task to process events
//static void ICACHE_FLASH_ATTR
static void uart_task(os_event_t *events)
{
    switch(events->sig) {
        case UART_SIG_ERR_FRM: {
            uart_debugSend("UART RX Frame error\r\n");
        }
        break;

        case UART_SIG_ERR_PARITY: {
            uart_debugSend("UART RX Parity error\r\n");
        }
        break;

        case UART_SIG_RXOVF: {
            uart_debugSend("UART RX overflow\r\n");
        }
        case UART_SIG_RECV: {
            static int bytesRecv = 0;

            //Set the activity LED
            _ledSet = 1;

            //Echo data
            int count = uart_get((char*)_rxBuffer, UART_RX_BUFFER_SIZE);

            bytesRecv += count;
            
            if(count > 0) {
                tcp_send((uint8*)_rxBuffer, count);

                uartCount = 0;

            }
        }
    break;


        
        case UART_SIG_TXTO: {
            //Transmit FIFO near empty

            //Grab as much as we can send
            uint8 sendSpace = uart_getTxFifoAvail();
            uint8 toSend = tcp_receive(_txBuffer, sendSpace);

            if(toSend > 0) {
                //Fill the FIFO
                uart0_send_nowait(_txBuffer, toSend);

                uart_set_txto();
                _ledSet = 1;
            }
            else {
                uart_clear_txto();

                _uartTxFlag = 0;
            }
        }
        break;

    default:
      break;
    }
}

void wifi_init() {
    char ssid[32];
    char password[64];
    
    //Set to Station mode
    wifi_set_opmode(STATION_MODE);

    //Set station mac
    //NOTE THIS MUST BE CALLED AFTER wifi_set_opmode
    //NOTE THIS CAN ONLY GO IN THE user_init function (wifi_init is called from user_init so it is okay)
    wifi_set_macaddr(STATION_IF, mac);

    //Station configuration 
    struct station_config stationConf;

    //Set ap settings
    strcpy(&stationConf.ssid, AP_SSID);
    strcpy(&stationConf.password, AP_PSK);

    //Set configuration
    wifi_station_set_config(&stationConf);

    //Set WiFi event handler
    wifi_set_event_handler_cb(&wifi_handler);
}
    
void wifi_handler(System_Event_t *evt) {
    char msg[256];
    switch(evt->event) {
#if DEBUG_WIFI == 1
        case EVENT_STAMODE_GOT_IP:
            os_sprintf(msg, "ip:" IPSTR ", mask:" IPSTR ",gw:" IPSTR "\n\r",
                    IP2STR(&evt->event_info.got_ip.ip),
                    IP2STR(&evt->event_info.got_ip.mask),
                    IP2STR(&evt->event_info.got_ip.gw));
            uart_debugSend(msg);
            break;
        case EVENT_STAMODE_DISCONNECTED:
            os_sprintf(msg, "disconnected from ssid: %s, reason %d\n\r",
                    evt->event_info.disconnected.ssid,
                    evt->event_info.disconnected.reason);
            uart_debugSend(msg);
            break;
        case EVENT_STAMODE_CONNECTED:
            os_sprintf(msg, "connect to ssid %s, channel %d\n\r", 
                    evt->event_info.connected.ssid, 
                    evt->event_info.connected.channel);
            uart_debugSend(msg);
            break;
#endif
        default:
            break;
    }
}

void tcp_recvHandler(uint16 len) {
    if(_uartTxFlag == 0) {
        //Grab as much as we can currently send
        uint8 sendSpace = uart_getTxFifoAvail();
        uint8 toSend = tcp_receive(_txBuffer, sendSpace);

        uart0_send_nowait(_txBuffer, toSend);
        uart_set_txto();

        _uartTxFlag = 1;
        _ledSet = 1;
    }
}

//Init function 
void ICACHE_FLASH_ATTR
user_init()
{
    //Remove debug statements from UART0
    system_set_os_print(0);

    // Initialize the GPIO subsystem.
    gpio_init();

    //Set GPIO2 to output mode
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_GPIO2_U, FUNC_GPIO2);

    //Set GPIO2 low
    gpio_output_set(0, BIT0, BIT0, 0);

    //Initialize UART
    uart_init(BAUD, BAUD);

    _rxBuffer = (uint8*)os_malloc(UART_RX_BUFFER_SIZE);
    _txBuffer = (uint8*)os_malloc(UART_TX_BUFFER_SIZE);

    _uartTxFlag = 0;

    //Initialize WiFi
    wifi_init();

#if DEBUG_WIFI == 1
    char mac_address[32];
    wifi_get_macaddr(0, &mac_address[0]);
    char msg[256];
    os_sprintf(msg, "MAC Address %02x-%02x-%02x-%02x-%02x-%02x\n\r",
            (unsigned) mac_address[0], (unsigned) mac_address[1],
            (unsigned) mac_address[2], (unsigned) mac_address[3],
            (unsigned) mac_address[4], (unsigned) mac_address[5]);
    uart_debugSend(msg);
#endif

    //Start TCP server
    tcp_start(TCP_PORT);
    tcp_setRecvHandler(&tcp_recvHandler);

    _ledSet = 0;

    //Disarm timer
    os_timer_disarm(&networkTimer);

    //Setup timer
    os_timer_setfn(&networkTimer, (os_timer_func_t *)network_task, NULL);

    //Arm the timer
    os_timer_arm(&networkTimer, NETWORK_UPDATE_RATE, 1);
    
    //Start os task
    system_os_task(uart_task, UART_TASK_PRIORITY,user_procTaskQueue,
            user_procTaskQueueLen);
}
